using EconPDEs, Test


println("Running tests:")

include("../examples/CampbellCochrane.jl")
try
	m = CampbellCochraneModel()
	state = initialize_state(m; n = 1000)
	y0 = initialize_y(m, state)
	y, a, distance = pdesolve(m, state, y0)
	@test distance <= 1e-5
	println("\t\033[1m\033[32mPASSED\033[0m: CampbellCochrane")
catch e
	println("\t\033[1m\033[31mFAILED\033[0m: CampbellCochrane")
	showerror(stdout, e, backtrace())
	rethrow(e)
end

include("../examples/BansalYaron.jl")
try
	m = BansalYaronModel()
	state = initialize_state(m; μn = 5, vn = 5)
	y0 = initialize_y(m, state)
	y, a, distance = pdesolve(m, state, y0)
	@test distance <= 1e-5
	println("\t\033[1m\033[32mPASSED\033[0m: BansalYaron")
catch e
	println("\t\033[1m\033[31mFAILED\033[0m: BansalYaron")
	showerror(stdout, e, backtrace())
	rethrow(e)
end


include("../examples/GarleanuPanageas.jl")
try
	m = GarleanuPanageasModel()
	state = initialize_state(m; n = 10)
	y0 = initialize_y(m, state)
	y, a, distance = pdesolve(m, state, y0)
	@test distance <= 1e-5
	println("\t\033[1m\033[32mPASSED\033[0m: GarleanuPanageas")
catch e
	println("\t\033[1m\033[31mFAILED\033[0m: GarleanuPanageas")
	showerror(stdout, e, backtrace())
	rethrow(e)
end


include("../examples/DiTella.jl")
try
	m = DiTellaModel()
	state = initialize_state(m ; xn = 10, νn = 3)
	y0 = initialize_y(m, state)
	y, a, distance = pdesolve(m, state, y0)
	@time pdesolve(m, state, y0, is_algebraic = OrderedDict(:pA => false, :pB => false, :p => true))
	@test distance <= 1e-5
	println("\t\033[1m\033[32mPASSED\033[0m: DiTella")
catch e
	println("\t\033[1m\033[31mFAILED\033[0m: DiTella")
	showerror(stdout, e, backtrace())
	rethrow(e)
end

include("../examples/WangWangYang.jl")
try
	m = WangWangYangModel()
	state = initialize_state(m; n = 10)
	y0 = initialize_y(m, state)
	y, a, distance = pdesolve(m, state, y0)
	@test distance <= 1e-5
	println("\t\033[1m\033[32mPASSED\033[0m: WangWangYang")
catch e
	println("\t\033[1m\033[31mFAILED\033[0m: WangWangYang")
	showerror(stdout, e, backtrace())
	rethrow(e)
end
